(* A first pass at compiling proofnet diagrams *)
(* type tree = Proofnettypes.tree
type perm = Proofnettypes.perm
type proofnet = Proofnettypes.proofnet*)

open Proofnettypes
open List

exception InvalidPermutation

(* I betray my java heritage! *)
let toString = string_of_int


(* print out some preamble *)
let do_header () =
  print_string "input boxes\n chunk = 1cm;\n"

  



(* ------------- generate new names as needed ------------ *)
let atomnumber = ref 0
let prevAtom () = "atom" ^ (toString (!atomnumber - 1))
let currAtom () = "atom" ^ (toString !atomnumber)
let newAtom () = ( atomnumber := !atomnumber + 1; "atom" ^ (toString !atomnumber))
let getAtom n = "atom" ^ (toString n)
let resetAtoms () = atomnumber := 0

let binopnumber = ref 0
let prevBinop () = "binop" ^ (toString (!binopnumber - 1))
let currBinop () = "binop" ^ (toString !binopnumber)
let newBinop () = ( binopnumber := !binopnumber + 1; "binop" ^ (toString !binopnumber))
let resetBinops () = binopnumber := 0
(*------------------------------------------------------------*)


(* ------- Ugly metapost code for calculating the points -------- *)

let make_atom_point atom = match atom with 
  "atom0" -> "(0,0)"
| _ as prev -> 
    "((xpart " ^ prev ^ ".c) + 2*chunk, ypart " ^ prev ^".c)"

let make_binop_point left right = 
  "((xpart " ^ left ^ ".c + xpart "^right^".c)/2, min(ypart "^left^".c, ypart "^right^".c) - chunk)"

let make_ax_link_midpoint left right height = 
  "((xpart " ^ left ^ ".c + xpart "^right^".c)/2, ypart "^left^".c + (("^height^" +1) * chunk/2))"
(*-------------------------------------------------*)




(* ----------------------------------------------------
 *  Return the metapost name assigned to this node; 
 *  generate metapost to draw the tree as a side-effect
 *  also increments the atom and binop counters as it draws the nodes
 *-----------------------------------------------------*)
let rec doTree = function 
    Atom s -> 
      let name = newAtom () in 
	print_string ("boxit." ^ name ^ "( btex $" ^ s ^ "$ etex);\n") ;
	print_string ( name ^ ".c = " ^ (make_atom_point (prevAtom ())) ^ ";\n") ; 
	print_string ( "drawunboxed(" ^ name ^ ");\n\n") ;
	name
  | Binop (l, r, s) -> 
      let left = doTree l in
      let right = doTree r in
      let name = newBinop () in 
       print_string ("circleit." ^ name ^ "( btex $" ^ s ^ "$ etex);\n") ;
       print_string ( name ^ ".c = " ^ (make_binop_point left right) ^ ";\n") ; 
       print_string ( "drawboxed(" ^ name ^ ");\n") ; 
       print_string ("draw "^left^".s{down}...{right}"^name^".w;\n");
       print_string ("draw "^right^".s{down}...{left}"^name^".e;\n\n");  
       name

(*----------------------------------------------------------------------*)
(* Draw in the axiom links, relying on the earlier defined mp names *)

(* this fun guarantees that:
   * l < r
   * the perm list is strictly increasing in left components
*)
let regularisePerms ps = 
  let rec validatePerms num found = function 
      [] -> []
    | (l,r)::t -> 
	if l <= num  && 
	  r <= num &&
	  not (mem l found) &&
	  not (mem r (l::found))
	then 
	  if l < r then
	    (l,r) :: (validatePerms num (r::l::found) t)
	  else
	    (r,l) :: (validatePerms num (r::l::found) t)
	else raise InvalidPermutation
  in
  sort (fun (l1,r1) (l2,r2) -> compare l1 l2) (validatePerms !atomnumber [] ps)


(* n current height, phs list of already heightified perms *)
let rec heightifyPerms n = function
    [] -> []
  | (l,r)::ps -> 
      let 
	contains (l1,r1) (l2,r2) = if l1 < l2 && r1 > r2 then true else false 
      in
      let (ps1, ps2) = partition (contains (l,r)) ps
      in 
      (l,r,n) :: ( (heightifyPerms (n+1) ps1) @ (heightifyPerms n ps2))

(* drawPerm generates the metapost code for the axiom link *)
let drawPerm (l,r,h) = 
  let left = getAtom l in
  let right = getAtom r in 
  let mpoint = make_ax_link_midpoint left right (string_of_int h) in 
  print_string ("draw " ^ left ^ ".n{up}...{right}"^ mpoint ^"{right}...{down}" ^ right ^ ".n;\n")


(*----------------------------------------------------------------------*)



(* handle tree part and perm part separately *)
let main (f,p) = 
  ignore (map doTree f ) ; 
  map drawPerm (heightifyPerms 1 (regularisePerms p))


(* each proof net generates a separate metapost figure *)
let drawFigure n pnet = 
  resetAtoms () ; resetBinops ();
  print_string ("\n beginfig(" ^ (toString n) ^ ");\n\n") ;
  ignore (main pnet) ;
  print_string "\n endfig; \n\n"


(* void main (void) *)
let _ = 
  let count = ref 1 in
  let drawnet pn = 
    drawFigure !count pn; count := !count + 1
  in
  let pnets = 
    Parser.proofnets Lexer.token (Lexing.from_channel stdin) 
  in
  do_header () ; 
  map drawnet pnets
